package admin

import (
	"context"
	"errors"
	"git.wamblee.org/converge/pkg/models"
	"log"
	"net"
	"sync"
)

// Extension of agentlistener for testing. It can accept all connections and puts them into a map based
// on clientId after which a client can retrieve the accepted connection based on client id.

type TestAgentListener struct {
	net.Listener

	ctx         context.Context
	mutex       sync.Mutex
	cond        *sync.Cond
	connections map[models.ClientId]net.Conn
}

func NewTestListener(ctx context.Context, listener net.Listener) *TestAgentListener {
	res := &TestAgentListener{
		ctx:         ctx,
		Listener:    listener,
		mutex:       sync.Mutex{},
		connections: make(map[models.ClientId]net.Conn),
	}
	res.cond = sync.NewCond(&res.mutex)

	go func() {
		for {
			conn, err := res.Accept()
			log.Printf("testlistener: Got connection %v %v", conn, err)
			if err != nil {
				return
			}
		}
	}()

	go func() {
		select {
		case <-res.ctx.Done():
			res.mutex.Lock()
			res.cond.Broadcast()
			res.mutex.Unlock()
		}
	}()
	return res
}

func (l *TestAgentListener) Accept() (net.Conn, error) {
	conn, err := l.Listener.Accept()
	if err != nil {
		return nil, err
	}
	clientId := models.ClientId(conn.LocalAddr().String())
	log.Printf("testlistener: Storing connection %v %v", clientId, conn)
	l.mutex.Lock()
	defer l.mutex.Unlock()
	l.connections[clientId] = conn
	log.Printf("testlistener: broadcasting %v", clientId)
	l.cond.Broadcast()
	return conn, err
}

func (l *TestAgentListener) getConnection(clientId models.ClientId) (net.Conn, error) {
	l.mutex.Lock()
	defer l.mutex.Unlock()
	// We need to check the condition before the first cond.wait as well. Otherwise, a broadcast sent
	// at this point in time will not be caught, and if there are no further broadcasts happening, then
	// the code will hang her.
	for ok := l.connections[clientId] != nil; !ok; ok = l.connections[clientId] != nil {
		log.Println("Listener cond wait")
		l.cond.Wait()
		log.Println("Listener awoken")
		select {
		case <-l.ctx.Done():
			return nil, errors.New("Listenere terminated because context canceled")
		default:
		}
	}
	log.Printf("Returning connection %v %v", clientId, l.connections[clientId])
	return l.connections[clientId], nil
}
